# Author: Alexandre Boucaud <aboucaud@lal.in2p3.fr>
# License: BSD 3 clause

import numpy as np

from keras.utils import multi_gpu_model
from keras.optimizers import Adam
from keras.callbacks import ModelCheckpoint, EarlyStopping, ReduceLROnPlateau, TensorBoard

from .score import iou
from .config import init_model_params
from .model import unet_dropout


class ObjectDetector(object):
    """Object detector.

    Parameters
    ----------
    batch_size : int, optional
        The batch size used during training. Set by default to 32 samples.

    epoch : int, optional
        The number of epoch for which the model will be trained. Set by default
        to 50 epochs.

    num_gpus : int, optional
        The number of gpus that will be used.

    model_check_point : bool, optional
        Whether to create a callback for intermediate models.

    Attributes
    ----------
    model_ : object
        The DNN model.

    params_model_ : Bunch dictionary
        All hyper-parameters to build the DNN model.

    """

    def __init__(self, batch_size=32, num_gpus=1, epoch=10, model_check_point=True,
                 filename=None, seed=42, plot_history=False,
                 output_path=None, bench_name=None):
        self.num_gpus = num_gpus
        self.model_, self.params_model_ = self._build_model()
        self.batch_size = batch_size
        self.epoch = epoch
        self.model_check_point = model_check_point
        self.filename = filename
        self.seed = seed
        self._plot_history = plot_history

        self.output_path = output_path
        self.bench_name = bench_name

    def fit(self, X, y):
        if X.ndim == 3:
            X = np.expand_dims(X, -1)
        if y.ndim == 3:
            y = np.expand_dims(y, -1)

        n_samples = len(X)

        id_val = int(n_samples * 0.8)

        X_train, X_val = np.split(X, [id_val])
        y_train, y_val = np.split(y, [id_val])

        # create the callbacks to get during fitting
        callbacks = self._build_callbacks()

        history = self.model_.fit(X_train, y_train,
                                  batch_size=self.batch_size, epochs=self.epoch,
                                  validation_data=(X_val, y_val),
                                  shuffle=True, callbacks=callbacks, verbose=1)

        if self._plot_history:
            self.plot_history(history)

    def predict(self, X):
        if X.ndim == 3:
            X = np.expand_dims(X, -1)
        return self.model_.predict(X)

    def predict_score(self, X, y_true):
        if X.ndim == 3:
            X = np.expand_dims(X, -1)
        if y_true.ndim == 3:
            y_true = np.expand_dims(y_true, -1)

        y_pred = self.model_.predict(X)
        return iou(y_true, y_pred)

    def plot_history(self, history):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt

        plt.semilogy(history.epoch, history.history['loss'], label='loss')
        plt.semilogy(
            history.epoch, history.history['val_loss'], label='val_loss')
        plt.title('Training performance')
        plt.legend()
        plt.savefig(
            "{0}/{1}_train_history.png".format(self.output_path, self.bench_name))
        plt.close()

    ###########################################################################
    # Setup model
    def _build_model(self):

        params_model = init_model_params()

        model = unet_dropout(
            input_shape=(params_model.img_rows,
                         params_model.img_cols,
                         params_model.img_channels),
            output_channels=params_model.output_channels,
            dropout=params_model.dropout_rate)

        optimizer = Adam(lr=params_model.lr,
                         beta_1=params_model.beta_1,
                         beta_2=params_model.beta_2,
                         epsilon=params_model.epsilon,
                         decay=params_model.decay)

        if self.num_gpus > 1:
            model = multi_gpu_model(model, gpus=self.num_gpus)

        model.compile(optimizer=optimizer,
                      loss=params_model.keras_loss,
                      metrics=['acc'])

        return model, params_model

    def _build_callbacks(self):
        callbacks = []

        if self.model_check_point:
            callbacks.append(
                ModelCheckpoint('{0}/{1}_unet_weights-best.h5'.format(self.output_path, self.bench_name),
                                monitor='val_loss',
                                save_best_only=True,
                                save_weights_only=True,
                                period=1,
                                verbose=1))
        # add early stopping
        if self.params_model_.early_stopping:
            callbacks.append(
                EarlyStopping(monitor='val_loss',
                              min_delta=self.params_model_.es_min_delta,
                              patience=self.params_model_.es_patience,
                              verbose=1))

        # reduce learning-rate when reaching plateau
        if self.params_model_.reduce_learning_rate:
            callbacks.append(
                ReduceLROnPlateau(monitor='val_loss',
                                  factor=self.params_model_.lr_factor,
                                  patience=self.params_model_.lr_patience,
                                  cooldown=self.params_model_.lr_cooldown,
                                  verbose=1))

        tensorboard = TensorBoard(log_dir=self.output_path, histogram_freq=0, write_graph=True, write_grads=False, write_images=False)
        callbacks.append(tensorboard)

        return callbacks
